@@ -10,6 +10,7 @@ import json
 import math
 from pathlib import Path
 import random
+import re
 from typing import Callable, Collection, Dict, Iterable, List, Optional, Set, Tuple, Union
 import uuid
 import xml.etree.ElementTree as ET
@@ -19,7 +20,7 @@ import networkx as nx
 import numpy as np
 import pinocchio as pin
 
-from timor import Robot
+from timor import Geometry, Robot
 from timor.Bodies import Body, BodyBase, BodySet, Connector, ConnectorSet, Gender
 from timor.Joints import Joint, JointSet, TimorJointType
 from timor.utilities import logging, spatial, write_urdf
@@ -138,6 +139,11 @@ class ModuleBase(abc.ABC):
         """Returns all joints contained in this module"""
         return self._joints
 
+    @property
+    def mass(self) -> float:
+        """Returns the total mass of the module."""
+        return sum(b.mass for b in self.bodies)
+
     @property
     def num_bodies(self):
         """The number of bodies in this module."""
@@ -232,6 +238,14 @@ class ModuleBase(abc.ABC):
             'joints': [joint.to_json_data() for joint in sorted(self.joints, key=lambda j: j.id)]
         }
 
+    def to_json_string(self) -> str:
+        """
+        Write a json in that fully describes this module.
+
+        :return: Returns the module specification in a json-ready dictionary
+        """
+        return json.dumps(self.to_json_data(), indent=2)
+
 
 class AtomicModule(ModuleBase):
     """
@@ -278,9 +292,9 @@ class AtomicModule(ModuleBase):
         return module
 
     @classmethod
-    def from_json_data(cls, d: Dict, package_dir: Path) -> 'ModuleBase':
+    def from_json_data(cls, d: Dict, package_dir: Path) -> AtomicModule:
         """
-        Maps the a json module description to an instance of this class.
+        Maps the json module description to an instance of this class.
 
         The dictionary will be modified in-place until empty (everything was parsed).
 
@@ -311,6 +325,17 @@ class AtomicModule(ModuleBase):
 
         return cls(header, bodies, joints)
 
+    @classmethod
+    def from_json_string(cls, s: str, package_dir: Path) -> AtomicModule:
+        """
+        Maps the json module string to an instance of this class.
+
+        :param package_dir: Package directory relative to which mesh file paths are defined
+        :param s: A json string with relevant meta-information
+        :return: An instantiated module
+        """
+        return cls.from_json_data(json.loads(s), package_dir)
+
 
 class ModulesDB(SingleSet):
     """
@@ -362,7 +387,7 @@ class ModulesDB(SingleSet):
         super().add(element)
 
     @classmethod
-    def from_file(cls, filepath: Path, package_dir: Path) -> 'ModulesDB':
+    def from_file(cls, filepath: Union[Path, str], package_dir: Union[Path, str]) -> ModulesDB:
         """
         Loads a modules Database from a json file.
 
@@ -370,40 +395,44 @@ class ModulesDB(SingleSet):
         :param package_dir: Package directory relative to which mesh file paths are defined
         :return: A ModulesDB
         """
+        filepath, package_dir = map(Path, (filepath, package_dir))
         with filepath.open('r') as json_file:
-            content = json.load(json_file)
-        return cls.from_json(content, package_dir)
+            content = json_file.read()
+        return cls.from_json_string(content, package_dir)
 
     @classmethod
-    def from_json(cls, json_content: Dict, package_dir: Path) -> 'ModulesDB':
+    def from_json_string(cls, json_string: str, package_dir: Path) -> ModulesDB:
         """
-        Loads a modules Database from a json file.
+        Loads a modules Database from a json string.
 
-        :param json_content: Json content of the modulesDB definition.
+        :param json_string: Json string of the modulesDB definition.
         :param package_dir: Package directory relative to which mesh file paths are defined
         :return: A ModulesDB
         """
         db = cls()
 
-        for module_data in json_content:
+        for module_data in json.loads(json_string):
             db.add(AtomicModule.from_json_data(module_data, package_dir))
         return db
 
-    def to_json(self, save_at: Path = None) -> str:
+    def to_json_file(self, save_at: Union[Path, str]):
         """
         Writes the ModulesDB to a json file.
 
-        :param save_at: If given, the file will be written to that path.
+        :param save_at: File location to write the DB to. Should be the "name" of the DB.
+        """
+        save_at = Path(save_at)
+        content = compress_json_vectors(self.to_json_string())
+        with Path(save_at).open('w') as savefile:
+            savefile.write(content)
+
+    def to_json_string(self) -> str:
+        """
+        Writes the ModulesDB to a json string.
+
         :return: The json string
         """
-        content = json.dumps([mod.to_json_data() for mod in sorted(self, key=lambda mod: mod.id)], indent=2)
-
-        content = compress_json_vectors(content)
-
-        if save_at is not None:
-            with save_at.open('w') as savefile:
-                savefile.write(content)
-        return content
+        return json.dumps([mod.to_json_data() for mod in sorted(self, key=lambda mod: mod.id)], indent=2)
 
     @property
     def all_module_names(self) -> Set[str]:
@@ -775,6 +804,11 @@ class ModuleAssembly:
             G.add_edge(mod_a.id, mod_b.id, label=[con_a._id, con_b._id])
         return G
 
+    @property
+    def mass(self) -> float:
+        """The total mass of the assembly"""
+        return sum(mod.mass for mod in self.module_instances)
+
     @property
     def nJoints(self) -> int:
         """The number of joints in the assembly"""
@@ -1009,8 +1043,14 @@ class ModuleAssembly:
 
         return robot
 
-    def to_urdf(self, name: str = None, write_to: Optional[Path] = None) -> str:
-        """Creates a URDF file from the assembly"""
+    def to_urdf(self, name: str = None, write_to: Optional[Path] = None, replace_wrl: bool = False) -> str:
+        """
+        Creates a URDF file from the assembly.
+
+        :param name: name to be given to robot inside URDF
+        :param write_to: output URDF file
+        :param replace_wrl: Whether to replace wrl geometries with visual geometry (if available).
+        """
         G = self.assembly_graph
         edges = self.assembly_graph.edges
         if name is None:
@@ -1037,10 +1077,28 @@ class ModuleAssembly:
                     body_name = '.'.join(successor.id)
                     link = ET.Element('link', {'name': body_name})
                     link.append(write_urdf.from_pin_inertia(successor.inertia, transform))
-                    link.append(write_urdf.from_geometry(successor.visual, 'visual',
-                                                         name=body_name + '_visual', link_frame=transform))
-                    link.append(write_urdf.from_geometry(successor.collision, 'collision',
-                                                         name=body_name + '_visual', link_frame=transform))
+
+                    visual_geom = None
+                    if not isinstance(successor.visual, Geometry.EmptyGeometry):
+                        visual_geom = write_urdf.from_geometry(successor.visual, 'visual',
+                                                               name=body_name + '_visual', link_frame=transform)
+                        link.append(visual_geom)
+
+                    if not isinstance(successor.collision, Geometry.EmptyGeometry):
+                        coll_geom = write_urdf.from_geometry(successor.collision, 'collision',
+                                                             name=body_name + '_collision', link_frame=transform)
+                        if replace_wrl:
+                            # Find all wrl references and replace with geometry found in visual description
+                            if visual_geom is None:
+                                raise ValueError("Cannot replace wrl if there is no visual geometry to replace with.")
+                            for m in coll_geom.findall('geometry/mesh'):
+                                if re.search(r"\.wrl$", m.get("filename")) is not None:
+                                    coll_geom = ET.Element("collision", {"name": coll_geom.get("name")})
+                                    for child in visual_geom:
+                                        coll_geom.append(child)
+                                    logging.info(f"Replaced geometry with wrl file by visual geom: {m.get('filename')}")
+                        link.append(coll_geom)
+
                     links.append(link)
                 elif isinstance(successor, Joint) and successor.type is not TimorJointType.fixed:
                     joint = ET.Element('joint', {'name': '.'.join(successor.id),
